<?php

namespace Drupal\simple_entity_merge;

use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Helper methods for merging entities.
 */
class SimpleEntityMerge {

  use DependencySerializationTrait;
  use StringTranslationTrait;

  /**
   * Drupal\Core\Entity\EntityFieldManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;
  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a new SimpleEntityMerge object.
   */
  public function __construct(EntityFieldManagerInterface $entity_field_manager, EntityTypeManagerInterface $entity_type_manager, MessengerInterface $messenger) {
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
  }

  /**
   * Merge references.
   *
   * Replace all references to the first entity id with references to the second
   * entity id.
   *
   * @param string $entity_type_id
   *   Entity type.
   * @param int $entity_source_id
   *   Source id.
   * @param int $entity_destination_id
   *   Destination id.
   */
  public function mergeReferences(string $entity_type_id, int $entity_source_id, int $entity_destination_id) {
    if (empty($entity_type_id) || empty($entity_source_id)) {
      return FALSE;
    }

    $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($entity_source_id);
    if (!$entity instanceof EntityInterface) {
      return FALSE;
    }

    // Create batch operations array.
    $operations = [];
    $batch_id = 0;

    $fields = $this->entityTypeManager->getStorage('field_storage_config')->loadMultiple();
    foreach ($fields as $field_name => $field) {
      // The field is entity reference and the target is of the same type as
      // the entity we want to merge.
      if ($field->getType() == 'entity_reference' && $field->getSetting('target_type') == $entity->getEntityTypeId()) {
        $query = $this->entityTypeManager->getStorage($field->getTargetEntityTypeId())->getQuery();
        $query->condition($field->getName(), $entity->id());
        $query->accessCheck(FALSE);
        $ids = $query->execute();

        // Split IDs into chunks of 50.
        $chunks = array_chunk($ids, 50);
        foreach ($chunks as $chunk) {
          $operations[] = [
            [$this, 'processFieldReferences'],
            [
              $field->getTargetEntityTypeId(),
              $chunk,
              $field->getName(),
              $entity->id(),
              $entity_destination_id,
              $batch_id++,
            ],
          ];
        }
      }
    }

    // For base fields we do much the same thing, looking, for instance, for
    // nodes owned by a user we are merging.
    $all_entity_types = $this->entityTypeManager->getDefinitions();
    foreach ($all_entity_types as $key => $type) {
      // Config entities have no base fields.
      if ($type instanceof ContentEntityType) {
        $revision_keys = !empty($type->getRevisionMetadataKeys()) ? array_values($type->getRevisionMetadataKeys()) : [];
        $defs = $this->entityFieldManager->getBaseFieldDefinitions($key);
        foreach ($defs as $field_name => $base_field_definition) {
          if ($base_field_definition->getType() === 'entity_reference' && $base_field_definition->getSetting('target_type') == $entity->getEntityTypeId() && !in_array($field_name, $revision_keys)) {
            $baseQuery = $this->entityTypeManager->getStorage($base_field_definition->getTargetEntityTypeId())->getQuery();
            $baseQuery->condition($base_field_definition->getName(), $entity->id());
            $baseQuery->accessCheck(FALSE);
            $ids = $baseQuery->execute();

            // Split IDs into chunks of 50.
            $chunks = array_chunk($ids, 50);
            foreach ($chunks as $chunk) {
              $operations[] = [
                [$this, 'processFieldReferences'],
                [
                  $base_field_definition->getTargetEntityTypeId(),
                  $chunk,
                  $base_field_definition->getName(),
                  $entity->id(),
                  $entity_destination_id,
                  $batch_id++,
                ],
              ];
            }
          }
        }
      }
    }

    // Set up and process the batch.
    $batch = [
      'title' => $this->t('Merging entity references...'),
      'operations' => $operations,
      'finished' => [$this, 'mergeReferencesFinished'],
    ];
    batch_set($batch);
    return TRUE;
  }

  /**
   * Batch operation callback to process field references.
   */
  public function processFieldReferences($entity_type_id, $entity_ids, $field_name, $source_id, $destination_id, $batch_id, &$context) {
    if (!isset($context['results']['processed'])) {
      $context['results']['processed'] = 0;
    }

    $entities = $this->entityTypeManager->getStorage($entity_type_id)->loadMultiple($entity_ids);
    foreach ($entities as $referencing_entity) {
      foreach ($referencing_entity->{$field_name} as $reference) {
        if ($reference->target_id == $source_id) {
          $reference->target_id = $destination_id;
        }
      }
      $referencing_entity->save();
      $context['results']['processed']++;
    }

    $context['message'] = $this->t('Processed @count entities', ['@count' => $context['results']['processed']]);
  }

  /**
   * Batch finished callback.
   */
  public function mergeReferencesFinished($success, $results, $operations) {
    if ($success) {
      if (!empty($results['processed'])) {
        $this->messenger->addMessage($this->t('Successfully processed @count entities.', ['@count' => $results['processed']]));
      }
      else {
        $this->messenger->addMessage($this->t('There were no entities to process.'));
      }
    }
    else {
      $this->messenger->addError($this->t('An error occurred while processing the entities.'));
    }
  }

}
